@@ -1,7 +1,6 @@
 {-# LANGUAGE BangPatterns       #-}
 {-# LANGUAGE DeriveDataTypeable #-}
 {-# LANGUAGE FlexibleContexts   #-}
-{-# LANGUAGE ImpredicativeTypes #-}
 {-# LANGUAGE OverloadedStrings  #-}
 {-# LANGUAGE PatternGuards      #-}
 {-# LANGUAGE RankNTypes         #-}
@@ -652,27 +651,38 @@ tag :: MonadThrow m
                                       --   of a tag, given the value return from the @AttrParser@
     -> CI.ConduitM Event o m (Maybe c)
 tag checkName attrParser f = do
-    x <- dropWS
-    case x of
+    (x, leftovers) <- dropWS []
+    res <- case x of
         Just (EventBeginElement name as) ->
             case checkName name of
                 Just y ->
                     case runAttrParser' (attrParser y) as of
                         Left e -> lift $ monadThrow e
                         Right z -> do
-                            CL.drop 1
                             z' <- f z
-                            a <- dropWS
+                            (a, _leftovers') <- dropWS []
                             case a of
                                 Just (EventEndElement name')
-                                    | name == name' -> CL.drop 1 >> return (Just z')
+                                    | name == name' -> return (Just z')
                                 _ -> lift $ monadThrow $ XmlException ("Expected end tag for: " ++ show name) a
                 Nothing -> return Nothing
         _ -> return Nothing
+
+    case res of
+        -- Did not parse, put back all of the leading whitespace events and the
+        -- final observed event generated by dropWS
+        Nothing -> mapM_ leftover leftovers
+        -- Parse succeeded, discard all of those whitespace events and the
+        -- first parsed event
+        Just _ -> return ()
+
+    return res
   where
-    -- Drop Events until we encount a non-whitespace element
-    dropWS = do
-        x <- CL.peek
+    -- Drop Events until we encounter a non-whitespace element. Return all of
+    -- the events consumed here (including the first non-whitespace event) so
+    -- that the calling function can treat them as leftovers if the parse fails
+    dropWS leftovers = do
+        x <- await
         let isWS =
                 case x of
                     Just EventBeginDocument -> True
@@ -689,7 +699,10 @@ tag checkName attrParser f = do
                     Just EventComment{} -> True
                     Just EventCDATA{} -> False
                     Nothing -> False
-        if isWS then CL.drop 1 >> dropWS else return x
+            leftovers' = maybe id (:) x leftovers
+        if isWS
+            then dropWS leftovers'
+            else return (x, leftovers')
     runAttrParser' p as =
         case runAttrParser p as of
             Left e -> Left e
@@ -816,9 +829,9 @@ orE a b =
 -- | Get the value of the first parser which returns 'Just'. If no parsers
 -- succeed (i.e., return 'Just'), this function returns 'Nothing'.
 choose :: Monad m
-       => [Consumer Event m (Maybe a)] -- ^ List of parsers that will be tried in order.
-       -> Consumer Event m (Maybe a) -- ^ Result of the first parser to succeed, or @Nothing@
-                                     --   if no parser succeeded
+       => [ConduitM Event o m (Maybe a)] -- ^ List of parsers that will be tried in order.
+       -> ConduitM Event o m (Maybe a)   -- ^ Result of the first parser to succeed, or @Nothing@
+                                         --   if no parser succeeded
 choose [] = return Nothing
 choose (i:is) =
     i >>= maybe (choose is) (return . Just)
